# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2.

from __future__ import absolute_import, division, print_function, unicode_literals

import logging

from typing import Any, cast, Dict, List, Optional, Union

import numpy as np
import pandas as pd
import scipy
import sklearn.utils.extmath
import statsmodels.api as sm

from balance import adjustment as balance_adjustment, util as balance_util
from balance.stats_and_plots.weights_stats import design_effect
from scipy.sparse import csc_matrix

logger: logging.Logger = logging.getLogger(__package__)


def logit_truncated(
    X: Union[np.ndarray, pd.DataFrame], beta: np.ndarray, truncation_value: float = 1e-5
) -> np.ndarray:
    """This is a helper function for cbps.
    Given an X matrx and avector of coeeficients beta, it computes the truncated
    version of the logit function.

    Args:
        X (Union[np.ndarray, pd.DataFrame]): Covariate matrix
        beta (np.ndarray): vector of coefficients
        truncation_value (float, optional): upper and lower bound for the computed probabilities. Defaults to 1e-5.

    Returns:
        np.ndarray: numpy array of computed probablities
    """
    probs = 1.0 / (1 + np.exp(-1 * (np.matmul(X, beta))))
    return np.minimum(np.maximum(probs, truncation_value), 1 - truncation_value)


def compute_pseudo_weights_from_logit_probs(
    probs: np.ndarray,
    design_weights: Union[np.ndarray, pd.DataFrame],
    in_pop: Union[np.ndarray, pd.DataFrame],
) -> np.ndarray:
    """This is a helper function for cbps.
    Given computed probs, it computes the weights: N/N_t * (in_pop - p_i)/(1 - p_i).
    (Note that these weights on sample are negative for convenience of notations)

    Args:
        probs (np.ndarray): vector of probabilities
        design_weights (Union[np.ndarray, pd.DataFrame]): vector of design weights of sample and target
        in_pop (Union[np.ndarray, pd.DataFrame]): indicator vector for target

    Returns:
        np.ndarray: np.ndarray of computed weights
    """
    N = np.sum(design_weights)
    N_target = np.sum(design_weights[in_pop == 1.0])
    return N / N_target * (in_pop - probs) / (1 - probs)


def bal_loss(
    beta: np.ndarray,
    X: np.ndarray,
    design_weights: np.ndarray,
    in_pop: np.ndarray,
    XtXinv: np.ndarray,
) -> np.float64:
    """This is a helper function for cbps.
    It computes the balance loss.

    Args:
        beta (np.ndarray): vector of coefficients
        X (np.ndarray): Covariate matrix
        design_weights (np.ndarray): vector of design weights of sample and target
        in_pop (np.ndarray): indicator vector for target
        XtXinv (np.ndarray): (X.T %*% X)^(-1)

    Returns:
        np.float64: computed balance loss
    """
    probs = logit_truncated(X, beta)
    N = np.sum(design_weights)
    weights = (
        1.0 / N * compute_pseudo_weights_from_logit_probs(probs, design_weights, in_pop)
    )

    Xprimew = np.matmul((X * design_weights[:, None]).T, weights)
    loss = np.absolute(np.matmul(np.matmul(Xprimew.T, XtXinv), Xprimew))

    return loss


def gmm_function(
    beta: np.ndarray,
    X: Union[np.ndarray, pd.DataFrame],
    design_weights: Union[np.ndarray, pd.DataFrame],
    in_pop: Union[np.ndarray, pd.DataFrame],
    invV: Union[np.ndarray, None] = None,
) -> Dict[str, Union[float, np.ndarray]]:
    """This is a helper function for cbps.
    It computes the gmm loss.

    Args:
        beta (np.ndarray): vector of coefficients
        X (Union[np.ndarray, pd.DataFrame]): covariates matrix
        design_weights (Union[np.ndarray, pd.DataFrame]): vector of design weights of sample and target
        in_pop (Union[np.ndarray, pd.DataFrame]): indicator vector for target
        invV (Union[np.ndarray, None], optional): the inverse weighting matrix for GMM. Default is None.

    Returns:
        Dict[str, Union[float, np.ndarray]]: Dict with two items for loss and invV:
            loss (float) computed gmm loss
            invV (np.ndarray) the weighting matrix for GMM
    """
    probs = logit_truncated(X, beta)
    N = np.sum(design_weights)
    N_target = np.sum(design_weights[in_pop == 1.0])

    weights = compute_pseudo_weights_from_logit_probs(probs, design_weights, in_pop)

    # Generate gbar
    gbar = np.concatenate(
        (
            1.0 / N * (np.matmul((X * design_weights[:, None]).T, (in_pop - probs))),
            1.0 / N * (np.matmul((X * design_weights[:, None]).T, weights)),
        )
    )

    if invV is None:
        # Compute inverse sigma matrix to use in GMM estimate
        design_weights_sq = np.sqrt(design_weights)[:, None]
        X1 = design_weights_sq * X * np.sqrt((1 - probs) * probs)[:, None]
        X2 = design_weights_sq * X * np.sqrt(probs / (1 - probs))[:, None]
        X11 = design_weights_sq * X * np.sqrt(probs)[:, None]
        X11TX11 = np.matmul(X11.T, X11 * N / N_target)

        V = (
            1
            / N
            * np.vstack(
                (
                    np.hstack((np.matmul(X1.T, X1), X11TX11)),
                    np.hstack(
                        (X11TX11, np.matmul(X2.T, X2 * np.power(N / N_target, 2)))
                    ),
                )
            )
        )
        # Note - The R CBPS code is using sum(treat) for N_target instead of
        # the sum of the design weights on the treated
        invV = np.linalg.pinv(V)

    # Compute loss
    loss = np.matmul(np.matmul(gbar.T, invV), gbar)
    return {"loss": loss, "invV": invV}


def gmm_loss(
    beta: np.ndarray,
    X: Union[np.ndarray, pd.DataFrame],
    design_weights: Union[np.ndarray, pd.DataFrame],
    in_pop: Union[np.ndarray, pd.DataFrame],
    invV: Optional[np.ndarray] = None,
) -> Union[float, np.ndarray]:
    """This is a helper function for cbps.
    It computes the gmm loss.
    See gmm_function for detials.

    Args:
        beta (np.ndarray): vector of coefficients
        X (Union[np.ndarray, pd.DataFrame]): covariates matrix
        design_weights (Union[np.ndarray, pd.DataFrame]): vector of design weights of sample and target
        in_pop (Union[np.ndarray, pd.DataFrame]): indicator vector for target
        invV (Union[np.ndarray, None], optional): the inverse weighting matrix for GMM. Default is None.

    Returns:
        Union[float, np.ndarray]: loss (float) computed gmm loss
    """
    return gmm_function(beta, X, design_weights, in_pop, invV)["loss"]


def alpha_function(
    alpha: np.ndarray,
    beta: np.ndarray,
    X: Union[np.ndarray, pd.DataFrame],
    design_weights: Union[np.ndarray, pd.DataFrame],
    in_pop: Union[np.ndarray, pd.DataFrame],
) -> Union[float, np.ndarray]:
    """This is a helper function for cbps.
    It computes the gmm loss of alpha*beta.

    Args:
        alpha (np.ndarray): multiplication factor
        beta (np.ndarray): vector of coefficients
        X (Union[np.ndarray, pd.DataFrame]): covariates matrix
        design_weights (Union[np.ndarray, pd.DataFrame]): vector of design weights of sample and target
        in_pop (Union[np.ndarray, pd.DataFrame]): indicator vector for target

    Returns:
        Union[float, np.ndarray]: loss (float) computed gmm loss
    """
    return gmm_loss(alpha * beta, X, design_weights, in_pop)


def compute_deff_from_beta(
    X: np.ndarray, beta: np.ndarray, design_weights: np.ndarray, in_pop: np.ndarray
) -> np.float64:
    """This is a helper function for cbps. It computes the design effect of
    the estimated weights on the sample given a value of beta.
    It is used for setting a constraints on max_de.

    Args:
        X (np.ndarray): covariates matrix
        beta (np.ndarray): vector of coefficients
        design_weights (np.ndarray): vector of design weights of sample and target
        in_pop (np.ndarray): indicator vector for target

    Returns:
        np.float64: design effect
    """
    probs = logit_truncated(X, beta)
    weights = np.absolute(
        compute_pseudo_weights_from_logit_probs(probs, design_weights, in_pop)
    )
    weights = design_weights[in_pop == 0.0] * weights[in_pop == 0.0]
    return design_effect(weights)


def _standardize_model_matrix(
    model_matrix: pd.DataFrame, model_matrix_columns_names: List[str]
) -> Dict[str, Any]:
    """This is a helper function for cbps. It standardizes the columns of the model matrix.

    Args:
        model_matrix (pd.DataFrame): the matrix of covariates
        model_matrix_columns_names (List[str]): list of columns in the covariates matrix

    Returns:
        Dict[str, Any]: Dict of the shape
            {
                "model_matrix": model_matrix,
                "model_matrix_columns_names": model_matrix_columns_names,
                "model_matrix_mean": model_matrix_mean,
                "model_matrix_std": model_matrix_std,
            }
    """
    # TODO: Verify if scikit-learn have something similar
    model_matrix_std = np.asarray(np.std(model_matrix, axis=0)).reshape(
        -1
    )  # This is needed if the input is 2 dim numpy array
    if np.sum(model_matrix_std == 0) > 0:
        variables_to_omit = model_matrix_std == 0
        names_variables_to_omit = np.array(model_matrix_columns_names)[
            variables_to_omit
        ]
        logger.warning(
            f"The following variables have only one level, and are omitted: {names_variables_to_omit}"
        )
        model_matrix_columns_names = list(
            np.array(model_matrix_columns_names)[np.invert(variables_to_omit)]
        )
        model_matrix = model_matrix[:, np.invert(variables_to_omit)]
        model_matrix_std = model_matrix_std[np.invert(variables_to_omit)]
    model_matrix_mean = np.mean(model_matrix, axis=0)
    model_matrix = (model_matrix - model_matrix_mean) / model_matrix_std
    return {
        "model_matrix": model_matrix,
        "model_matrix_columns_names": model_matrix_columns_names,
        "model_matrix_mean": model_matrix_mean,
        "model_matrix_std": model_matrix_std,
    }


# TODO: update docs
def _reverse_svd_and_centralization(beta, U, s, Vh, X_matrix_mean, X_matrix_std):
    """This is a helper function for cbps. It revrse the svd and the centralization to get an estimate of beta
    Source: https://github.com/kosukeimai/CBPS/blob/master/R/CBPSMain.R#L353

    Args:
        beta (_type_): _description_
        U (_type_): _description_
        s (_type_): _description_
        Vh (_type_): _description_
        X_matrix_mean (_type_): _description_
        X_matrix_std (_type_): _description_

    Returns:
        _type_: _description_
    """
    # TODO: update SVD and reverse SVD to the functions in scikit-learn
    # Invert s
    s_inv = s
    s_inv[s_inv > 1e-5] = 1 / s_inv[s_inv > 1e-5]
    s_inv[s_inv <= 1e-5] = 0
    # Compute beta
    beta = np.matmul(Vh.T * s_inv, beta)
    beta_new = np.delete(beta, 0) / X_matrix_std
    beta_new = np.insert(
        beta_new,
        0,
        beta[
            0,
        ]
        - np.matmul(X_matrix_mean, beta_new),
    )
    return beta_new


def cbps(  # noqa
    sample_df: pd.DataFrame,
    sample_weights: pd.Series,
    target_df: pd.DataFrame,
    target_weights: pd.Series,
    variables: Optional[List[str]] = None,
    transformations: str = "default",
    na_action: str = "add_indicator",
    formula: Optional[Union[str, List[str]]] = None,
    balance_classes: bool = True,
    cbps_method: str = "over",  # other option: "exact"
    max_de: Optional[float] = None,
    opt_method: str = "COBYLA",
    opt_opts: Optional[Dict] = None,
    weight_trimming_mean_ratio: Union[None, float, int] = 20,
    weight_trimming_percentile: Optional[float] = None,
    random_seed: int = 2020,
    *args,
    **kwargs,
) -> Dict[str, Union[pd.Series, Dict]]:
    """Fit cbps (covariate balancing propensity score model) for the sample using the target.
    Final weights are normalized to target size.
    We use a two-step GMM estimator (as in the default R package), unlike the suggeted continuous-updating
    estimator in the paper. The reason is that it runs much faster than the continuous one.

    Paper: Imai, K., & Ratkovic, M. (2014). Covariate balancing propensity score.
    Journal of the Royal Statistical Society: Series B: Statistical Methodology, 243-263.
    https://rss.onlinelibrary.wiley.com/doi/abs/10.1111/rssb.12027
    R code source: https://github.com/kosukeimai/CBPS
    two-step GMM: https://en.wikipedia.org/wiki/Generalized_method_of_moments

    Args:
        sample_df (pd.DataFrame): a dataframe representing the sample
        sample_weights (pd.Series): design weights for sample
        target_df (pd.DataFrame): a dataframe representing the target
        target_weights (pd.Series): design weights for target
        variables (Optional[List[str]], optional): list of variables to include in the model.
            If None all joint variables of sample_df and target_df are used. Defaults to None.
        transformations (str, optional): what transformations to apply to data before fitting the model.
            Default is "default" (see apply_transformations function). Defaults to "default".
        na_action (str, optional): what to do with NAs. (see add_na_indicator function).
            Defaults to "add_indicator".
        formula (Optional[Union[str, List[str]]], optional): The formula according to which build the model.
            In case of list of formula, the model matrix will be built in steps and
            concatenated together.. Defaults to None.
        balance_classes (bool, optional): whether to balance the sample and target size for running the model.
            True is preferable for imbalanced cases. Defaults to True.
        cbps_method (str, optional): method used for cbps. "over" fits an over-identified model that combines
            the propensity score and covariate balancing conditions; "exact" fits a model that only c
            ontains the covariate balancing conditions. Defaults to "over".
        max_de (Optional[float], optional): upper bound for the design effect of the computed weights.
            Default is None.
        opt_method (str, optional): type of optimization solver. See :func:`scipy.optimize.minimize`
                   for other options. Defaults to "COBYLA".
        opt_opts (Optional[Dict], optional): A dictionary of solver options. Default is None. See :func:`scipy.optimize.minimize`
            for other options. Defaults to None.
        weight_trimming_mean_ratio (Union[None, float, int], optional): indicating the ratio from above according to which
            the weights are trimmed by mean(weights) * ratio. Defaults to 20.
        weight_trimming_percentile (Optional[float], optional): if weight_trimming_percentile is not none, winsorization is applied.
            Default is None, i.e. trimming is applied.
        random_seed (int, optional): a random seed. Defaults to 2020.

    Raises:
        Exception: _description_
        Exception: _description_
        Exception: _description_
        Exception: _description_

    Returns:
        Dict[str, Union[pd.Series, Dict]]: A dictionary includes:
        "weight" --- The weights for the sample.
        "model" -- dictionary with details about the fitted model:
            X_matrix_columns, deviance, beta_optimal, balance_optimize_result,
            gmm_optimize_result_glm_init, gmm_optimize_result_bal_init
            It has the following shape:
            "model": {
                "method": "cbps",
                "X_matrix_columns": X_matrix_columns_names,
                "deviance": deviance,
                "original_sum_weights": original_sum_weights,  # This can be used to reconstruct the propensity probablities
                "beta_optimal": beta_opt,
                "beta_init_glm": beta_0,  # The initial estimator by glm
                "gmm_init": gmm_init,  # The rescaled initial estimator
                # The following are the results of the optimizations
                "rescale_initial_result": rescale_initial_result,
                "balance_optimize_result": balance_optimize_result,
                "gmm_optimize_result_glm_init": gmm_optimize_result_glm_init
                if cbps_method == "over"
                else None,
                "gmm_optimize_result_bal_init": gmm_optimize_result_bal_init
                if cbps_method == "over"
                else None,
            },
    """

    logger.info("Starting cbps function")
    np.random.seed(random_seed)  # setting random seed for cases of variations in glmnet

    balance_util._check_weighting_methods_input(sample_df, sample_weights, "sample")
    balance_util._check_weighting_methods_input(target_df, target_weights, "target")

    # Choose joint variables from sample and target
    variables = balance_util.choose_variables(sample_df, target_df, variables=variables)
    logger.debug(f"Joint variables for sample and target: {variables}")

    sample_df = sample_df.loc[:, variables]
    target_df = target_df.loc[:, variables]

    if na_action == "drop":
        (sample_df, sample_weights) = balance_util.drop_na_rows(
            sample_df, sample_weights, "sample"
        )
        (target_df, target_weights) = balance_util.drop_na_rows(
            target_df, target_weights, "target"
        )
    # keeping index of sample df to use for final weights
    sample_index = sample_df.index

    # Applying transformations
    # Important! Variables that don't need transformations
    # should be transformed with the *identity function*,
    # otherwise will be dropped from the model
    sample_df, target_df = balance_adjustment.apply_transformations(
        (sample_df, target_df), transformations=transformations
    )
    variables = list(sample_df.columns)
    logger.debug(f"Final variables in the model: {variables}")

    # Build X matrix
    model_matrix_output = balance_util.model_matrix(
        sample_df,
        target_df,
        variables,
        add_na=(na_action == "add_indicator"),
        return_type="one",
        return_var_type="sparse",
        # pyre-fixme[6]: for 7th parameter `formula` expected `Optional[List[str]]` but got `Union[None, List[str], str]`.
        formula=formula,
        one_hot_encoding=False,
    )
    # TODO: Currently using a dense version of the X matrix. We might change to using the sparse version if need.
    X_matrix = cast(
        Union[pd.DataFrame, np.ndarray, csc_matrix],
        (model_matrix_output["model_matrix"]),
    ).toarray()
    X_matrix_columns_names = model_matrix_output["model_matrix_columns_names"]
    logger.info(
        f"The formula used to build the model matrix: {model_matrix_output['formula']}"
    )

    # Standardize the X_matrix for SVD
    model_matrix_standardized = _standardize_model_matrix(
        X_matrix,
        # pyre-fixme[6]: for 2nd positional only parameter expected `List[str]` but got `Union[None, List[str], ndarray, DataFrame, csc_matrix]`.
        X_matrix_columns_names,
    )
    X_matrix = model_matrix_standardized["model_matrix"]
    X_matrix_columns_names = model_matrix_standardized["model_matrix_columns_names"]
    logger.info(f"The number of columns in the model matrix: {X_matrix.shape[1]}")
    logger.info(f"The number of rows in the model matrix: {X_matrix.shape[0]}")

    # Adding intercept since model_matrix removes it
    X_matrix = np.c_[np.ones(X_matrix.shape[0]), X_matrix]
    X_matrix_columns_names.insert(0, "Intercept")

    # SVD for X_matrix
    U, s, Vh = scipy.linalg.svd(X_matrix, full_matrices=False)
    # Make the sign of the SVD deterministic
    U, Vh = sklearn.utils.extmath.svd_flip(U, Vh, u_based_decision=False)
    # TODO: add stop: if (k < ncol(X)) stop("X is not full rank")

    sample_n = sample_df.shape[0]
    target_n = target_df.shape[0]
    total_n = sample_n + target_n

    # Create "treatment" (in_sample) variable
    in_sample = np.concatenate((np.ones(sample_n), np.zeros(target_n)))
    in_pop = np.concatenate((np.zeros(sample_n), np.ones(target_n)))
    if len(np.unique(in_sample.reshape(in_sample.shape[0]))) == 1:
        _number_unique = np.unique(in_sample.reshape(in_sample.shape[0]))
        raise Exception(
            f"Sample indicator only has value {_number_unique}. This can happen when your sample or target are empty from unknown reason"
        )

    # balance classes
    if balance_classes:
        design_weights = (
            total_n
            / 2
            * np.concatenate(
                (
                    sample_weights / np.sum(sample_weights),
                    target_weights / np.sum(target_weights),
                )
            )
        )

        XtXinv = np.linalg.pinv(
            # pyre-fixme[16] Undefined attribute [16]: `float` has no attribute `__getitem__`.
            np.matmul((U * design_weights[:, None]).T, U * design_weights[:, None])
        )
    else:
        design_weights = np.concatenate((sample_weights, target_weights))
        design_weights = design_weights / np.mean(design_weights)
        XtXinv = np.linalg.pinv(np.matmul(U.T, U))

    # Define constraints for optimization to limit max_de
    constraints = []
    if max_de is not None:
        constraints += [
            # max de effect
            {
                "type": "ineq",
                "fun": lambda x: (
                    max_de - compute_deff_from_beta(U, x, design_weights, in_pop)
                ),
            }
        ]

    # Optimization using Generalized Methods of Moments:
    # Step 0 - initial estimation for beta
    logger.info("Finding initial estimator for GMM optimization")
    glm_mod = sm.GLM(
        in_pop, U, family=sm.families.Binomial(), freq_weights=design_weights
    )
    beta_0 = glm_mod.fit().params
    # TODO: add some safty for when this fails?

    # Step 1 - rescale initial estimation for beta by minimizing the gmm loss
    rescale_initial_result = scipy.optimize.minimize(
        alpha_function,
        x0=[1],
        args=(
            beta_0,
            U,
            design_weights,
            in_pop,
        ),
        bounds=[(0.8, 1.1)],  # These are the bounds used in the R CBPS package
    )
    if rescale_initial_result["success"] is np.bool_(False):
        logger.warning(
            f"Convergence of alpha_function has failed due to '{rescale_initial_result['message']}'"
        )
    gmm_init = beta_0 * rescale_initial_result["x"][0]
    invV = gmm_function(gmm_init, U, design_weights, in_pop, invV=None)["invV"]

    # Step 2 - find initial estimation for beta by minimizing balance loss
    logger.info(
        "Finding initial estimator for GMM optimization that minimizes the balance loss"
    )
    balance_optimize_result = scipy.optimize.minimize(
        fun=bal_loss,
        x0=gmm_init,
        args=(
            U,
            design_weights,
            in_pop,
            XtXinv,
        ),
        method=opt_method,
        options=opt_opts,
        constraints=constraints,
    )
    if balance_optimize_result["success"] is np.bool_(False):
        logger.warning(
            f"Convergence of bal_loss function has failed due to '{balance_optimize_result['message']}'"
        )
    beta_balance = balance_optimize_result["x"]

    gmm_optimize_result_glm_init = None
    if cbps_method == "exact":
        if (
            balance_optimize_result["success"] is np.bool_(False)
            and "Did not converge to a solution satisfying the constraints"
            in balance_optimize_result["message"]
        ):
            raise Exception("There is no solution satisfying the constraints.")
        beta_opt = beta_balance

    # Step 3 - minimize gmm_loss from two starting points: beta_balance and gmm_init,
    # and choose the solution that minimize the gmm_loss.
    elif cbps_method == "over":
        logger.info("Running GMM optimization")
        gmm_optimize_result_glm_init = scipy.optimize.minimize(
            fun=gmm_loss,
            x0=gmm_init,
            args=(
                U,
                design_weights,
                in_pop,
                invV,
            ),
            method=opt_method,
            options=opt_opts,
            constraints=constraints,
        )
        if gmm_optimize_result_glm_init["success"] is np.bool_(False):
            logger.warning(
                f"Convergence of gmm_loss function with gmm_init start point has failed due to '{gmm_optimize_result_glm_init['message']}'"
            )

        gmm_optimize_result_bal_init = scipy.optimize.minimize(
            fun=gmm_loss,
            x0=beta_balance,
            args=(
                U,
                design_weights,
                in_pop,
                invV,
            ),
            method=opt_method,
            options=opt_opts,
            constraints=constraints,
        )
        if gmm_optimize_result_bal_init["success"] is np.bool_(False):
            logger.warning(
                f"Convergence of gmm_loss function with beta_balance start point has failed due to '{gmm_optimize_result_bal_init['message']}'"
            )

        # If the constraints cannot be satisfied, exit the function
        if (
            gmm_optimize_result_glm_init["success"] is np.bool_(False)
            and "Did not converge to a solution satisfying the constraints"
            in gmm_optimize_result_glm_init["message"]
            and gmm_optimize_result_bal_init["success"] is np.bool_(False)
            and "Did not converge to a solution satisfying the constraints"
            in gmm_optimize_result_bal_init["message"]
        ):
            raise Exception("There is no solution satisfying the constraints.")

        # Choose beta that gives a smaller loss in GMM and that satisfy the constraints
        if (
            gmm_optimize_result_bal_init["fun"] < gmm_optimize_result_glm_init["fun"]
            and "Did not converge to a solution satisfying the constraints"
            not in gmm_optimize_result_bal_init["message"]
        ):
            beta_opt = gmm_optimize_result_bal_init["x"]
        else:
            beta_opt = gmm_optimize_result_glm_init["x"]

    else:
        raise Exception(f"cbps_method '{cbps_method}' is not a valid option")

    # Compute final probs and weights with beta_opt
    probs = logit_truncated(U, beta_opt)
    logger.debug(f"Minimum probs for sample: {np.min(probs[in_sample == 1.0])}")
    logger.debug(f"Maximum probs for sample: {np.max(probs[in_sample == 1.0])}")
    weights = np.absolute(
        compute_pseudo_weights_from_logit_probs(probs, design_weights, in_pop)
    )
    logger.debug(f"Minimum weight for sample: {np.min(weights[in_sample == 1.0])}")
    logger.debug(f"Maximum weight for sample: {np.max(weights[in_sample == 1.0])}")
    weights = design_weights[in_sample == 1.0] * weights[in_sample == 1.0]

    # Compute deviance (Minus twice the log-likelihood of the CBPS fit)
    deviance = -2 * np.sum(
        in_pop * design_weights * np.log(probs)
        + (1 - in_pop) * design_weights * np.log(1 - probs)
    )

    # trim weights
    weights = balance_adjustment.trim_weights(
        weights, weight_trimming_mean_ratio, weight_trimming_percentile
    )
    # normalize to target size
    original_sum_weights = np.sum(weights)
    logger.debug(f"original sum of weights for sample: {original_sum_weights}")
    weights = weights / original_sum_weights * np.sum(target_weights)

    # set index to sample_df
    weights = pd.DataFrame({"weight": weights}).set_index(sample_index)["weight"]

    if np.unique(weights).shape[0] == 1 or weights.describe()["std"] < 1e-4:
        # All weights are the same
        logger.warning(
            "All weights are identical (or almost identical). The estimates will not be adjusted"
        )

    # Revrse SVD and centralization
    beta_opt = _reverse_svd_and_centralization(
        beta_opt,
        U,
        s,
        Vh,
        model_matrix_standardized["model_matrix_mean"],
        model_matrix_standardized["model_matrix_std"],
    )

    out = {
        "weight": weights,
        "model": {
            "method": "cbps",
            "X_matrix_columns": X_matrix_columns_names,
            "deviance": deviance,
            "original_sum_weights": original_sum_weights,  # This can be used to reconstruct the propensity probablities
            "beta_optimal": beta_opt,
            "beta_init_glm": beta_0,  # The initial estimator by glm
            "gmm_init": gmm_init,  # The rescaled initial estimator
            # The following are the results of the optimizations
            "rescale_initial_result": rescale_initial_result,
            "balance_optimize_result": balance_optimize_result,
            "gmm_optimize_result_glm_init": gmm_optimize_result_glm_init
            if cbps_method == "over"
            else None,
            # pyre-fixme[61]: `gmm_optimize_result_bal_init` is undefined, or not
            #  always defined.
            "gmm_optimize_result_bal_init": gmm_optimize_result_bal_init
            if cbps_method == "over"
            else None,
        },
    }
    logger.info("Done cbps function")

    return out



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2.

from __future__ import absolute_import, division, print_function, unicode_literals

import logging

import random

from typing import Dict, List, Literal, Optional, Tuple, Union

import matplotlib.pyplot as plt

import numpy as np
import pandas as pd
import plotly.graph_objs as go
import plotly.offline as offline
import seaborn as sns

from balance.stats_and_plots.weighted_stats import (
    relative_frequency_table,
    weighted_quantile,
)
from balance.util import choose_variables, rm_mutual_nas
from matplotlib.colors import rgb2hex

logger: logging.Logger = logging.getLogger(__package__)


################################################################################
#  seaborn plots below
################################################################################


def _return_sample_palette(
    names: List[str],
) -> Union[Dict[str, str], str]:
    """Returns sample palette for seaborn plots.

    Named colors for matplotlib: https://stackoverflow.com/a/37232760

    Args:
        names (List[str]): e.g. ['self', 'target']

    Returns:
        Union[Dict[str, str], str]: e.g. {'self': 'tomato', 'target': 'dodgerblue'}
    """
    # colors to match the plotly colors
    col_unadjusted = rgb2hex((222 / 255, 45 / 255, 38 / 255, 0.8), keep_alpha=True)
    col_adjusted = rgb2hex((52 / 255, 165 / 255, 48 / 255, 0.5), keep_alpha=True)
    col_target = rgb2hex((158 / 255, 202 / 255, 225 / 255, 0.8), keep_alpha=True)

    if set(names) == {"self", "target"}:
        sample_palette = {
            "self": col_unadjusted,
            "target": col_target,
        }
    elif set(names) == {"self", "unadjusted"}:
        sample_palette = {
            "self": col_adjusted,
            "unadjusted": col_unadjusted,
        }
    elif set(names) == {"self", "unadjusted", "target"}:
        sample_palette = {
            "self": col_adjusted,
            "unadjusted": col_unadjusted,
            "target": col_target,
        }
    elif set(names) == {"adjusted", "unadjusted", "target"}:
        sample_palette = {
            "adjusted": col_adjusted,
            "unadjusted": col_unadjusted,
            "target": col_target,
        }
    else:
        sample_palette = "muted"
    return sample_palette


def _plotly_marker_color(
    name: Literal["sample", "unadjusted", "self", "adjusted"],
    only_self_and_target: bool,
    color_type: Literal["color", "line"],
) -> str:
    """
    Returns a color string for a marker in a plotly plot based on the given parameters.

    Args:
        name (Literal["sample", "unadjusted", "self", "adjusted"]): Name of the marker.
        only_self_and_target (bool): Determines if only self and target groups are available, or if it's self, unadjusted and target.
        color_type (Literal["color", "line"]): The type of color, either "color" or "line".

    Returns:
        str: A string representing the color in RGBA format.

    Raises:
        ValueError: If the color_type is not one of the accepted options.
    """
    if color_type == "color":
        col1 = 0.8
        col2 = 0.5
    elif color_type == "line":
        col1 = 1
        col2 = 1
    else:
        raise ValueError(
            "Invalid value for 'tycolor_typepe'. Must be either 'color' or 'line'."
        )

    if name.lower() in ["sample", "unadjusted"] or (
        name.lower() == "self" and only_self_and_target
    ):
        return f"rgba(222,45,38,{col1})"
    elif name.lower() in ["self", "adjusted"]:
        return f"rgba(52,165,48,{col2})"
    else:
        return f"rgb(158,202,225,{col1})"


def plot_bar(
    dfs: List[Dict[str, Union[pd.DataFrame, pd.Series]]],
    names: List[str],
    column: str,
    axis: Optional[plt.Axes] = None,
    weighted: bool = True,
    title: Optional[str] = None,
    ylim: Optional[Tuple[float, float]] = None,
) -> None:
    """
    Shows a (weighted) sns.barplot using a relative frequency table of several DataFrames (with optional control over the y-axis limits).

    If weighted is True, then mutual NA values are removed using :func:`rm_mutual_nas`.

    Args:
        dfs (List[Dict[str, Union[pd.DataFrame, pd.Series]]]): a list (of length 1 or more) of dictionaries which describe the DataFrames and weights
            The structure is as follows:
            [
                {'df': pd.DataFrame(...), "weight": pd.Series(...)},
                ...
            ]
            The 'df' is a DataFrame which includes the column name that was supplied through 'column'.
            The "weight" is a pd.Series of weights that are used when aggregating the variable using :func:`relative_frequency_table`.
        names (List[str]): a list of the names of the DataFrames that are plotted. E.g.: ['adjusted', 'unadjusted', 'target']
        column (str): The column to be used to aggregate using :func:`relative_frequency_table`.
        axis (Optional[plt.Axes], optional): matplotlib Axes object to draw the plot onto, otherwise uses the current Axes. Defaults to None.
        weighted (bool, optional): If to pass the weights from the dicts inside dfs. Defaults to True.
        title (str, optional): Title of the plot. Defaults to "barplot of covar '{column}'".
        ylim (Optional[Tuple[float, float]], optional): A tuple with two float values representing the lower and upper limits of the y-axis.
            If not provided, the y-axis range is determined automatically. Defaults to None.

    Examples:
        ::

            from balance.stats_and_plots.weighted_comparisons_plots import plot_bar
            import pandas as pd
            import numpy as np

            df = pd.DataFrame({
                'group': ('a', 'b', 'c', 'c'),
                'v1': (1, 2, 3, 4),
            })

            plot_bar(
                [{"df": df, "weight": pd.Series((1, 1, 1, 1))}, {"df": df, "weight": pd.Series((2, 1, 1, 1))}],
                names = ["self", "target"],
                column = "group",
                axis = None,
                weighted = True)

            # The same as above just with ylim set to (0, 1).
            plot_bar(
                [{"df": df, "weight": pd.Series((1, 1, 1, 1))}, {"df": df, "weight": pd.Series((2, 1, 1, 1))}],
                names = ["self", "target"],
                column = "group",
                axis = None,
                weighted = True,
                ylim = (0, 1))

            # Also deals with np.nan weights
            a = plot_bar(
                [{"df": df, "weight": pd.Series((1, 1, 1, np.nan))}, {"df": df, "weight": pd.Series((2, 1, 1, np.nan))}],
                names = ["self", "target"],
                column = "group",
                axis = None,
                weighted = True)
    """
    plot_data = []
    for ii, i in enumerate(dfs):
        a_series = i["df"][column]
        _w = i["weight"]
        if weighted:
            a_series, _w = rm_mutual_nas(a_series, _w)
            a_series.name = column  # rm_mutual_nas removes name, so we set it back

        df_plot_data = relative_frequency_table(a_series, w=_w if weighted else None)
        df_plot_data["dataset"] = names[ii]  # a recycled column for barplot's hue.

        plot_data.append(df_plot_data)

    sample_palette = _return_sample_palette(names)
    if title is None:
        title = f"barplot of covar '{column}'"

    ax = sns.barplot(
        x=column,
        y="prop",
        hue="dataset",
        data=pd.concat(plot_data),
        ax=axis,
        palette=sample_palette,
        saturation=1,
        alpha=0.6,
    )

    if ylim is not None:
        ax.set_ylim(ylim)

    ax.set_title(title)


def plot_hist_kde(
    dfs: List[Dict[str, Union[pd.DataFrame, pd.Series]]],
    names: List[str],
    column: str,
    axis: Optional[plt.Axes] = None,
    weighted: bool = True,
    dist_type: Literal["hist", "kde", "ecdf"] = "hist",
    title: Optional[str] = None,
) -> None:
    """Shows a (weighted) distribution plot ():func:`sns.displot`) of data from several DataFrame objects.

    Options include histogram (hist), kernel density estimate (kde), and empirical cumulative density function (ecdf).

    Args:
        dfs (List[Dict[str, Union[pd.DataFrame, pd.Series]]]): a list (of length 1 or more) of dictionaries which describe the DataFrames and weights
            The structure is as follows:
            [
                {'df': pd.DataFrame(...), "weight": pd.Series(...)},
                ...
            ]
            The 'df' is a DataFrame which includes the column name that was supplied through 'column'.
            The "weight" is a pd.Series of weights that are used when aggregating the variable using :func:`relative_frequency_table`.
        names (List[str]): a list of the names of the DataFrames that are plotted. E.g.: ['adjusted', 'unadjusted', 'target']
        column (str): The column to be used to aggregate using :func:`relative_frequency_table`.
        axis (Optional[plt.Axes], optional): matplotlib Axes object to draw the plot onto, otherwise uses the current Axes. Defaults to None.
        weighted (bool, optional): If to pass the weights from the dicts inside dfs. Defaults to True.
        dist_type (Literal["hist", "kde", "ecdf"], optional): The type of plot to draw. Defaults to "hist".
        title (str, optional): Title of the plot. Defaults to "distribution plot of covar '{column}'".


    Examples:
        ::

            from balance.stats_and_plots.weighted_comparisons_plots import plot_hist_kde
            import pandas as pd
            import numpy as np
            import matplotlib.pyplot as plt

            df = pd.DataFrame({
                'group': ('a', 'b', 'c', 'c'),
                'v1': (1, 2, 3, 4),
            })

            dfs1 = [{"df": pd.DataFrame(pd.Series([1,2,2,2,3,4,5,5,7,8,9,9,9,9,5,2,5,4,4,4], name = "v1")), "weight": None}, {"df": df, "weight": pd.Series((200, 1, 0, 20))}]

            plt.figure(1)

            # kde: no weights
            plot_hist_kde(
                dfs1,
                names = ["self", "target"],
                column = "v1",
                axis = None,
                weighted = False, dist_type = "kde")

            plt.figure(2)

            # kde: with weights
            plot_hist_kde(
                dfs1,
                names = ["self", "target"],
                column = "v1",
                axis = None,
                weighted = True, dist_type = "kde")

            plt.figure(3)

            # hist
            plot_hist_kde(
                dfs1,
                names = ["self", "target"],
                column = "v1",
                axis = None,
                weighted = True, dist_type = "hist")

            plt.figure(4)

            # ecdf
            plot_hist_kde(
                dfs1,
                names = ["self", "target"],
                column = "v1",
                axis = None,
                weighted = True, dist_type = "ecdf")


            # can work nicely with plt.subplots:
            f, axes = plt.subplots(1, 2, figsize=(7, 7 * 1))
            plot_hist_kde(
                dfs1,
                names = ["self", "target"],
                column = "v1",
                axis = axes[0],
                weighted = False, dist_type = "kde")
            plot_hist_kde(
                dfs1,
                names = ["self", "target"],
                column = "v1",
                axis = axes[1],
                weighted = False, dist_type = "kde")
    """
    possible_dist_function = {
        "hist": sns.histplot,
        "kde": sns.kdeplot,
        "ecdf": sns.ecdfplot,
    }
    dist_function = possible_dist_function[dist_type]
    # NOTE: the reason we don't use sns.displot directly is that it doesn't accept an ax= kwarg.
    # see also here: https://stackoverflow.com/a/63895570/256662

    plot_data = []
    for ii, i in enumerate(dfs):
        a_series = i["df"][column]
        _w = i["weight"]
        if _w is None:
            _w = pd.Series(np.ones(len(a_series)), index=a_series.index)
        if weighted:
            a_series, _w = rm_mutual_nas(a_series, _w)
            a_series.name = column  # rm_mutual_nas removes name, so we set it back
            # TODO: verify if this normalization to sum to 1 is needed (if so, how come we don't do it when _w is None)?
            _w = _w / np.sum(_w)

        df_plot_data = pd.DataFrame({column: a_series, "_w": _w})
        df_plot_data["dataset"] = names[ii]  # a recycled column for barplot's hue.

        plot_data.append(df_plot_data)

    plot_data = pd.concat(plot_data)

    sample_palette = _return_sample_palette(names)
    if title is None:
        title = f"distribution plot of covar '{column}'"

    kwargs_for_dist_function = {
        "data": plot_data,
        "x": column,
        "hue": "dataset",
        "ax": axis,
        "weights": plot_data["_w"] if weighted else None,
        # common_norm:False,
        "palette": sample_palette,
        "linewidth": 2,
    }
    if dist_type != "ecdf":
        kwargs_for_dist_function["common_norm"] = False
    ax = dist_function(**kwargs_for_dist_function)
    ax.set_title(title)


def plot_qq(
    dfs: List[Dict[str, Union[pd.DataFrame, pd.Series]]],
    names: List[str],
    column: str,
    axis: Optional[plt.Axes] = None,
    weighted: bool = True,
) -> None:
    """Plots a qq plot of the weighted data from a DataFrame object against some target.

    See: https://en.wikipedia.org/wiki/Q-Q_plot

    Args:
        dfs (List[Dict[str, Union[pd.DataFrame, pd.Series]]]): a list (of length 1 or more) of dictionaries which describe the DataFrames and weights
            The structure is as follows:
            [
                {'df': pd.DataFrame(...), "weight": pd.Series(...)},
                ...
            ]
            The 'df' is a DataFrame which includes the column name that was supplied through 'column'.
            The "weight" is a pd.Series of weights that are used when aggregating the variable using :func:`weighted_quantile`.
            Uses the last df item in the list as the target.
        names (List[str]): a list of the names of the DataFrames that are plotted. E.g.: ['adjusted', 'unadjusted', 'target']
        column (str): The column to be used to aggregate using :func:`weighted_quantile`.
        axis (Optional[plt.Axes], optional): matplotlib Axes object to draw the plot onto, otherwise uses the current Axes. Defaults to None.
        weighted (bool, optional): If to pass the weights from the dicts inside dfs. Defaults to True.

    Examples:
        ::

            import numpy as np
            import pandas as pd
            from balance.stats_and_plots.weighted_comparisons_plots import plot_qq
            from numpy import random

            df = pd.DataFrame({
                'v1': random.uniform(size=100),
            }).sort_values(by=['v1'])

            dfs1 = [
                {"df": df, "weight": pd.Series(np.ones(100))},
                {"df": df, "weight": pd.Series(range(100))},
                {"df": df, "weight": pd.Series(np.ones(100))},
            ]

            # plot_qq(dfs1, names=["self", "unadjusted", "target"], column="v1", axis=None, weighted=False)
            plot_qq(dfs1, names=["self", "unadjusted", "target"], column="v1", axis=None, weighted=True)
    """
    target = dfs[-1]  # assumes the last item is target
    dfs = dfs[:-1]  # assumes the non-last item are dfs to be compared to target

    for ii, i in enumerate(dfs):
        d = i["df"]
        _w = i["weight"]

        target_q = weighted_quantile(
            target["df"][column],
            np.arange(0, 1, 0.001),
            # pyre-fixme[6]: TODO:
            # This is because of using:
            # dfs: List[Dict[str, Union[pd.DataFrame, pd.Series]]],
            # When in fact we want to be clear that the first element is called
            # "df" and the second "weight", and that the first is a pd.DataFrame and
            # the second pd.Series. Until this is not clear - the following line will raise an error.
            target["weight"] if weighted else None,
        )
        sample_q = weighted_quantile(
            d.loc[:, column],
            np.arange(0, 1, 0.001),
            # pyre-fixme[6]
            _w if weighted else None,
        )

        axis = sns.scatterplot(
            x=target_q.iloc[:, 0].values,
            y=sample_q.iloc[:, 0].values,
            label=names[ii],
            ax=axis,
        )
    set_xy_axes_to_use_the_same_lim(axis)

    axis.plot(axis.get_xlim(), axis.get_ylim(), "--")
    axis.set_title(f"quantile-quantile plot of covar '{column}' in target vs sample")
    axis.legend()


def plot_qq_categorical(
    dfs: List[Dict[str, Union[pd.DataFrame, pd.Series]]],
    names: List[str],
    column: str,
    axis: Optional[plt.Axes] = None,
    weighted: bool = True,
    label_threshold: int = 30,
) -> None:
    """A scatter plot of weighted relative frequencies of categories from each df.

    Notice that this is not a "real" qq-plot, but rather a scatter plot of (estimated, weighted) probabilities for each category.

    X-axis is the sample (adjusted, unadjusted) and Y-axis is the target.

    Args:
        dfs (List[Dict[str, Union[pd.DataFrame, pd.Series]]]): a list (of length 1 or more) of dictionaries which describe the DataFrames and weights
            The structure is as follows:
            [
                {'df': pd.DataFrame(...), "weight": pd.Series(...)},
                ...
            ]
            The 'df' is a DataFrame which includes the column name that was supplied through 'column'.
            The "weight" is a pd.Series of weights that are used when aggregating the variable using :func:`weighted_quantile`.
            Uses the last df item in the list as the target.
        names (List[str]): a list of the names of the DataFrames that are plotted. E.g.: ['adjusted', 'unadjusted', 'target']
        column (str): The column to be used to aggregate using :func:`relative_frequency_table`.
        axis (Optional[plt.Axes], optional): matplotlib Axes object to draw the plot onto, otherwise uses the current Axes. Defaults to None.
        weighted (bool, optional): If to pass the weights from the dicts inside dfs. Defaults to True.
        label_threshold (int, optional): All labels that are larger from the threshold will be omitted from the scatter plot (so to reduce clutter). Defaults to 30.

    Examples:
        ::

            import numpy as np
            import pandas as pd
            from balance.stats_and_plots.weighted_comparisons_plots import plot_qq_categorical
            from numpy import random

            df = pd.DataFrame({
                'v1': random.random_integers(11111, 11114, size=100),
            }).sort_values(by=['v1'])

            dfs1 = [
                {"df": df, "weight": pd.Series(np.ones(100))},
                {"df": df, "weight": pd.Series(np.ones(99).tolist() + [1000])},
                {"df": df, "weight": pd.Series(np.ones(100))},
            ]

            import matplotlib.pyplot as plt

            plt.rcParams["figure.figsize"] = (20, 6) # (w, h)

            fig, axs = plt.subplots(1,3)

            # Without using weights
            plot_qq_categorical(dfs1, names=["self", "unadjusted", "target"], column="v1", axis=axs[0], weighted=False)
            # With weights
            plot_qq_categorical(dfs1, names=["self", "unadjusted", "target"], column="v1", axis=axs[1], weighted=True)
            # With label trimming if the text is longer than 3.
            plot_qq_categorical(dfs1, names=["self", "unadjusted", "target"], column="v1", axis=axs[2], weighted=True, label_threshold=3)
    """
    target = dfs[-1]["df"]
    target_weights = dfs[-1]["weight"]
    dfs = dfs[:-1]

    # pyre-fixme[6]
    target_plot_data = relative_frequency_table(target, column, target_weights)

    for ii, i in enumerate(dfs):
        d = i["df"]
        _w = i["weight"] if weighted else None

        # pyre-fixme[6]
        sample_plot_data = relative_frequency_table(d, column, _w)
        plot_data = pd.merge(
            sample_plot_data,
            target_plot_data,
            on=column,
            how="outer",
            suffixes=("_sample", "_target"),
        )

        axis = sns.scatterplot(
            x=plot_data.prop_sample, y=plot_data.prop_target, label=names[ii], ax=axis
        )

        if plot_data.shape[0] < label_threshold:
            for r in plot_data.itertuples():
                # pyre-fixme[16]: `tuple` has no attribute `prop_sample`.
                axis.text(x=r.prop_sample, y=r.prop_target, s=r[1])

    axis.set_ylim(-0.1, 1.1)
    axis.set_xlim(-0.1, 1.1)
    axis.plot(axis.get_xlim(), axis.get_ylim(), "--")
    axis.set_title(
        f"proportion-proportion plot of covar '{column}' in target vs sample"
    )
    axis.legend()


# TODO: add control (or just change) the default theme
# TODO: add a separate dist_type control parameter for categorical and numeric variables.
def seaborn_plot_dist(
    dfs: List[Dict[str, Union[pd.DataFrame, pd.Series]]],
    names: Optional[List[str]] = None,
    variables: Optional[List] = None,
    numeric_n_values_threshold: int = 15,
    weighted: bool = True,
    dist_type: Optional[Literal["qq", "hist", "kde", "ecdf"]] = None,
    return_axes: bool = False,
    ylim: Optional[Tuple[float, float]] = None,
) -> Union[List[plt.Axes], np.ndarray, None]:
    """Plots to compare the weighted distributions of an arbitrary number of variables from
    an arbitrary number of DataFrames.

    Uses: :func:`plot_qq_categorical`, :func:`plot_qq`, :func:`plot_hist_kde`, :func:`plot_bar`.

    Args:
        dfs (List[Dict[str, Union[pd.DataFrame, pd.Series]]]): a list (of length 1 or more) of dictionaries which describe the DataFrames and weights
            The structure is as follows:
            [
                {'df': pd.DataFrame(...), "weight": pd.Series(...)},
                ...
            ]
            The 'df' is a DataFrame which includes the column name that was supplied through 'column'.
            The "weight" is a pd.Series of weights that are used when aggregating by the column variable.
        names (List[str]): a list of the names of the DataFrames that are plotted. E.g.: ['adjusted', 'unadjusted', 'target']
        variables (Optional[List], optional): The list of variables to use, by default (None) will plot all of them.
        numeric_n_values_threshold (int, optional): How many unique values (or less) should be in a column so that it is considered to be a "category"? Defaults to 15.
            This is compared against the maximum number of distinct values (for each of the variables) across all DataFrames.
            Setting this value to 0 will disable this check.
        weighted (bool, optional): If to pass the weights from the dicts inside dfs. Defaults to True.
        dist_type (Optional[str], optional): can be "hist", "kde", or "qq". Defaults to None.
        return_axes (bool, optional): if to returns axes or None. Defaults to False,
        ylim (Optional[Tuple[float, float]], optional): A tuple with two float values representing the lower and upper limits of the y-axis.
            If not provided, the y-axis range is determined automatically. Defaults to None.
            Passed only for categorical variables and when dist_type is not 'qq' (i.e.: for bar plots).


    Returns:
        Union[List[plt.Axes], np.ndarray, None]: Returns None.
        However, if return_axes is True then either it returns a list or an np.array of matplotlib AxesSubplot (plt.Subplot).
        NOTE: There is no AxesSubplot class until one is invoked and created on the fly.
            See details here: https://stackoverflow.com/a/11690800/256662

    Examples:
        ::

            import numpy as np
            import pandas as pd
            from balance.stats_and_plots.weighted_comparisons_plots import seaborn_plot_dist
            from numpy import random

            df = pd.DataFrame({
                'v1': random.random_integers(11111, 11114, size=100).astype(str),
                'v2': random.normal(size = 100),
                'v3': random.uniform(size = 100),
            }).sort_values(by=['v2'])

            dfs1 = [
                {"df": df, "weight": pd.Series(np.ones(100))},
                {"df": df, "weight": pd.Series(np.ones(99).tolist() + [1000])},
                {"df": df, "weight": pd.Series(np.random.uniform(size=100))},
            ]

            seaborn_plot_dist(dfs1, names=["self", "unadjusted", "target"], dist_type = "qq")  # default
            seaborn_plot_dist(dfs1, names=["self", "unadjusted", "target"], dist_type = "hist")
            seaborn_plot_dist(dfs1, names=["self", "unadjusted", "target"], dist_type = "kde")
            seaborn_plot_dist(dfs1, names=["self", "unadjusted", "target"], dist_type = "ecdf")

            # With limiting the y axis range to (0,1)
            seaborn_plot_dist(dfs1, names=["self", "unadjusted", "target"], dist_type = "kde", ylim = (0,1))
    """
    if dist_type is None:
        if len(dfs) == 1:
            dist_type = "hist"
        else:
            dist_type = "qq"

    #  Choose set of variables to plot
    variables = choose_variables(*(d["df"] for d in dfs), variables=variables)
    logger.debug(f"plotting variables {variables}")

    #  Set up subplots
    f, axes = plt.subplots(len(variables), 1, figsize=(7, 7 * len(variables)))
    if not isinstance(axes, np.ndarray):  # If only one subplot
        axes = [axes]

    # TODO: patch choose_variables to return outcome_types from multiple_objects
    numeric_variables = dfs[0]["df"].select_dtypes(exclude=["object"]).columns.values

    for io, o in enumerate(variables):
        #  Find the maximum number of non-missing values of this variable accross
        #  all the dataframes
        n_values = max(len(set(rm_mutual_nas(d["df"].loc[:, o].values))) for d in dfs)
        if n_values == 0:
            logger.warning(f"No nonmissing values for variable '{o}', skipping")
            continue

        #  Plot categorical variables as histogram
        categorical = (o not in numeric_variables) or (
            n_values < numeric_n_values_threshold
        )

        if categorical:
            if dist_type == "qq":
                # pyre-fixme[6]
                plot_qq_categorical(dfs, names, o, axes[io], weighted)
            else:
                # pyre-fixme[6]
                plot_bar(dfs, names, o, axes[io], weighted, ylim=ylim)
        else:
            if dist_type == "qq":
                # pyre-fixme[6]
                plot_qq(dfs, names, o, axes[io], weighted)
            else:
                # pyre-fixme[6]
                plot_hist_kde(dfs, names, o, axes[io], weighted, dist_type)

    if return_axes:
        return axes
    # else (default) will return None


def set_xy_axes_to_use_the_same_lim(ax: plt.Axes) -> None:
    """Set the x and y axes limits to be the same.

    Done by taking the min and max from xlim and ylim and using these global min/max on both x and y axes.

    Args:
        ax (plt.Axes): matplotlib Axes object to draw the plot onto.

    Examples:
        ::

            import matplotlib.pyplot as plt
            plt.figure(1)
            plt.scatter(x= [1,2,3], y = [3,4,5])

            plt.figure(2)
            fig, ax = plt.subplots(1, 1, figsize=(7.2, 7.2))
            plt.scatter(x= [1,2,3], y = [3,4,5])
            set_xy_axes_to_use_the_same_lim(ax)
    """
    xlim = ax.get_xlim()
    ylim = ax.get_ylim()
    ax.set_xlim(min(xlim[0], ylim[0]), max(xlim[1], ylim[1]))
    ax.set_ylim(min(xlim[0], ylim[0]), max(xlim[1], ylim[1]))


################################################################################
#  plotly plots below
################################################################################


def plotly_plot_qq(
    dict_of_dfs: Dict[str, pd.DataFrame],
    variables: List[str],
    plot_it: bool = True,
    return_dict_of_figures: bool = False,
) -> Optional[Dict[str, go.Figure]]:
    """
    Plots interactive QQ plot of the given variables.

    Creates a plotly qq plot of the given variables from multiple DataFrames.
    This ASSUMES there is a df with key 'target'.


    Args:
        dict_of_dfs (Dict[str, pd.DataFrame]): The key is the name of the DataFrame (E.g.: self, unadjusted, target),
            and the value is the DataFrame that contains the variables that we want to plot.
        variables (List[str]): a list of variables to use for plotting.
        plot_it (bool, optional): If to plot the plots interactively instead of returning a dictionary. Defaults to True.
        return_dict_of_figures (bool, optional): If to return the dictionary containing the plots rather than just returning None. Defaults to False.

    Returns:
        Optional[Dict[str, go.Figure]]: Dictionary containing plots if return_dict_of_figures is True. None otherwise.

    Examples:
        ::

            import numpy as np
            import pandas as pd
            from numpy import random
            from balance.stats_and_plots.weighted_comparisons_plots import plotly_plot_qq

            random.seed(96483)

            df = pd.DataFrame({
                'v1': random.random_integers(11111, 11114, size=100).astype(str),
                'v2': random.normal(size = 100),
                'v3': random.uniform(size = 100),
            }).sort_values(by=['v2'])

            dict_of_dfs = {
                "self": pd.concat([df, pd.Series(random.random(size = 100) + 0.5, name = "weight")], axis = 1),
                "unadjusted": pd.concat([df, pd.Series(np.ones(99).tolist() + [1000], name = "weight")], axis = 1),
                "target": pd.concat([df, pd.Series(np.ones(100), name = "weight")], axis = 1),
            }

            # It won't work with "v1" since it is not numeric.
            plotly_plot_qq(dict_of_dfs, variables= ["v2", "v3"])
    """
    dict_of_qqs = {}
    for variable in variables:
        variable_specific_dict_of_plots = {}

        assert "target" in dict_of_dfs.keys(), "Must pass target"

        # Extract 'col1' because weighted_quantile will return a DataFrame
        # https://www.statsmodels.org/dev/_modules/statsmodels/stats/weightstats.html#DescrStatsW.quantile

        line_data = list(
            weighted_quantile(
                dict_of_dfs["target"][variable],
                np.arange(0, 1, 0.001),
                dict_of_dfs["target"]["weight"]
                if "weight" in dict_of_dfs["target"].columns
                else None,
            )["col1"]
        )

        # Indicate if we have only self and target (without unadjusted)
        # since in this case the color of self should be red, since it's likely unadjusted.
        only_self_and_target = set(dict_of_dfs.keys()) == {"self", "target"}

        for name in dict_of_dfs:
            if name.lower() == "target":
                variable_specific_dict_of_plots[name] = go.Scatter(
                    x=line_data,
                    y=line_data,
                    showlegend=False,
                    line={"color": ("blue"), "width": 2, "dash": "dash"},
                )
            else:
                variable_specific_dict_of_plots[name] = go.Scatter(
                    x=line_data,
                    y=list(
                        weighted_quantile(
                            dict_of_dfs[name][variable],
                            np.arange(0, 1, 0.001),
                            dict_of_dfs[name]["weight"]
                            if "weight" in dict_of_dfs[name].columns
                            else None,
                        )["col1"]
                    ),
                    marker={
                        "color": _plotly_marker_color(
                            # pyre-ignore[6]: it cannot get to this point if name=="target".
                            name,
                            only_self_and_target,
                            "color",
                        )
                    },
                    mode="markers",
                    name=naming_legend(name, list(dict_of_dfs.keys())),
                    opacity=0.6,
                )

        data = [variable_specific_dict_of_plots[name] for name in dict_of_dfs]
        layout = {
            "title": f"QQ Plots of {variable}",
            "paper_bgcolor": "rgb(255, 255, 255)",
            "plot_bgcolor": "rgb(255, 255, 255)",
            "xaxis": {
                "zeroline": False,
                "linewidth": 1,
                "mirror": True,
                "title": variable,
            },
            "yaxis": {"zeroline": False, "linewidth": 1, "mirror": True},
        }
        fig = go.Figure(data=data, layout=layout)
        dict_of_qqs[variable] = fig
        if plot_it:
            offline.iplot(fig)
    if return_dict_of_figures:
        return dict_of_qqs


def plotly_plot_density(
    dict_of_dfs: Dict[str, pd.DataFrame],
    variables: List[str],
    plot_it: bool = True,
    return_dict_of_figures: bool = False,
    plot_width: int = 800,
) -> Optional[Dict[str, go.Figure]]:
    """
    Plots interactive density plots of the given variables using kernel density estimation.

    Creates a plotly plot of the kernel density estimate for each variable in the given list
    across multiple DataFrames. The function assumes there is a DataFrame with the key 'target'.
    The density plot shows the distribution of the variable for each DataFrame in the dictionary.
    It looks for a `weights` column and uses it to normalize the data. If no weight column is found, it assumes all weights are equal to 1.
    It relies on the seaborn library to create the KDE (`sns.kdeplot`).

    Args:
        dict_of_dfs (Dict[str, pd.DataFrame]): A dictionary where each key is a name for the DataFrame
            and the value is the DataFrame that contains the variables to plot.
        variables (List[str]): A list of variables to plot.
        plot_it (bool, optional): Whether to plot the figures interactively using plotly. Defaults to True.
        return_dict_of_figures (bool, optional): Whether to return a dictionary of plotly figures.
            Defaults to False.
        plot_width (int, optional): The width of the plot in pixels. Defaults to 800.

    Returns:
        Optional[Dict[str, go.Figure]]: A dictionary containing plotly figures for each variable
        in the given list if `return_dict_of_figures` is True. Otherwise, returns None.

    Examples:
        ::
            import numpy as np
            import pandas as pd
            from numpy import random
            from balance.stats_and_plots.weighted_comparisons_plots import plotly_plot_density, plot_dist

            random.seed(96483)

            df = pd.DataFrame({
                'v1': random.random_integers(11111, 11114, size=100).astype(str),
                'v2': random.normal(size = 100),
                'v3': random.uniform(size = 100),
            }).sort_values(by=['v2'])

            dict_of_dfs = {
                "self": pd.concat([df, pd.Series(random.random(size = 100) + 0.5, name = "weight")], axis = 1),
                "unadjusted": pd.concat([df, pd.Series(np.ones(99).tolist() + [1000], name = "weight")], axis = 1),
                "target": pd.concat([df, pd.Series(np.ones(100), name = "weight")], axis = 1),
            }

            # It won't work with "v1" since it is not numeric.
            plotly_plot_density(dict_of_dfs, variables= ["v2", "v3"], plot_width = 550)


            # The above gives the same results as:
            dfs1 = [
                {"df": df, "weight": dict_of_dfs['self']["weight"]},
                {"df": df, "weight": dict_of_dfs['unadjusted']["weight"]},
                {"df": df, "weight": dict_of_dfs['target']["weight"]},
            ]
            plot_dist(dfs1, names=["self", "unadjusted", "target"], library="seaborn", dist_type = "kde", variables= ["v2", "v3"])


            # This gives the same shape of plots (notice how we must have the column "weight" for the plots to work)
            df = pd.DataFrame({
                'group': ('a', 'b', 'c', 'c'),
                'v1': (1, 2, 3, 4),
            })

            dfs1 = [{"df": pd.DataFrame(pd.Series([1,2,2,2,3,4,5,5,7,8,9,9,9,9,5,2,5,4,4,4], name = "v1")), "weight": None}, {"df": df, "weight": pd.Series((200, 1, 0, 200000))}]
            # dfs1[1]{'df'}

            dict_of_dfs = {
                "self": dfs1[0]['df'], # pd.concat([df, pd.Series(random.random(size = 100) + 0.5, name = "weight")], axis = 1),
                "target": pd.concat([dfs1[1]['df'], pd.Series(dfs1[1]["weight"], name = "weight")], axis = 1),
            }

            plotly_plot_density(dict_of_dfs, variables= ["v1"], plot_width = 550)

            plot_dist(dfs1, names=["self", "target"], library="seaborn", dist_type = "kde", variables= ["v1"],numeric_n_values_threshold = 1)

    """
    dict_of_density_plots = {}

    for variable in variables:
        data = []

        # Indicate if we have only self and target (without unadjusted)
        # since in this case the color of self should be red, since it's likely unadjusted.
        only_self_and_target = set(dict_of_dfs.keys()) == {"self", "target"}

        for name, df in dict_of_dfs.items():
            if "weight" in df.columns:
                weights = df["weight"]
                # TODO: verify if this normalization to sum to 1 is needed (if so, how come we don't do it when _w is None)?
                weights = weights / weights.sum()  # normalize weights by sum of weights
            else:
                weights = np.ones(len(df))

            # Convert the data to long format
            long_df = pd.DataFrame({"value": df[variable], "weight": weights})

            # Replace KDE calculation with sns.kdeplot
            with plt.xkcd():
                plt.figure()
                ax = sns.kdeplot(
                    data=long_df, x="value", weights="weight", common_norm=False
                )
                x = ax.get_lines()[-1].get_xdata()
                y = ax.get_lines()[-1].get_ydata()
                # print(name)
                # print(x)
                # print(y)
                plt.close()

            trace = go.Scatter(
                x=x,
                y=y,
                mode="lines",
                name=naming_legend(name, list(dict_of_dfs.keys())),
                line={
                    # pyre-ignore[6]: it cannot get to this point if name=="target".
                    "color": _plotly_marker_color(name, only_self_and_target, "line"),
                    "width": 1.5,
                },
            )
            data.append(trace)

        layout = {
            "title": f"Density Plots of '{variable}'",
            "paper_bgcolor": "rgb(255, 255, 255)",
            "plot_bgcolor": "rgb(255, 255, 255)",
            "width": plot_width,
            "xaxis": {
                "title": variable,
                "gridcolor": "rgba(128, 128, 128, 0.5)",
                "gridwidth": 1,
                "showgrid": True,
            },
            "yaxis": {
                "title": "Density",
                "gridcolor": "rgba(128, 128, 128, 0.5)",
                "gridwidth": 1,
                "showgrid": True,
            },
        }
        fig = go.Figure(data=data, layout=layout)
        dict_of_density_plots[variable] = fig

        if plot_it:
            offline.iplot(fig)

    if return_dict_of_figures:
        return dict_of_density_plots


def plotly_plot_bar(
    dict_of_dfs: Dict[str, pd.DataFrame],
    variables: List[str],
    plot_it: bool = True,
    return_dict_of_figures: bool = False,
    ylim: Optional[Tuple[float, float]] = None,
) -> Optional[Dict[str, go.Figure]]:
    """
    Plots interactive bar plots of the given variables (with optional control over the y-axis limits).

    Args:
        dict_of_dfs (Dict[str, pd.DataFrame]): A dictionary with keys as names of the DataFrame (e.g., 'self', 'unadjusted', 'target'),
            and values as the DataFrames containing the variables to plot.
        variables (List[str]): A list of variables to use for plotting.
        plot_it (bool, optional): If True, plots the graphs interactively instead of returning a dictionary. Defaults to True.
        return_dict_of_figures (bool, optional): If True, returns the dictionary containing the plots rather than just returning None. Defaults to False.
        ylim (Optional[Tuple[float, float]], optional): A tuple with two float values representing the lower and upper limits of the y-axis.
            If not provided, the y-axis range is determined automatically. Defaults to None.

    Returns:
        Optional[Dict[str, go.Figure]]: Dictionary containing plots if return_dict_of_figures is True. None otherwise.

    Examples:
        ::

            import numpy as np
            import pandas as pd
            from numpy import random
            from balance.stats_and_plots.weighted_comparisons_plots import plotly_plot_bar

            random.seed(96483)

            df = pd.DataFrame({
                'v1': random.random_integers(11111, 11114, size=100).astype(str),
                'v2': random.normal(size = 100),
                'v3': random.uniform(size = 100),
            }).sort_values(by=['v2'])

            dict_of_dfs = {
                "self": pd.concat([df, pd.Series(random.random(size = 100) + 0.5, name = "weight")], axis = 1),
                "unadjusted": pd.concat([df, pd.Series(np.ones(99).tolist() + [1000], name = "weight")], axis = 1),
                "target": pd.concat([df, pd.Series(np.ones(100), name = "weight")], axis = 1),
            }

            # It can work with "v2" and "v3", but it would be very sparse
            plotly_plot_bar(dict_of_dfs, variables= ["v1"])

            # Plots the same as above, but this time the range of the yaxis is from 0 to 1.
            plotly_plot_bar(dict_of_dfs, variables= ["v1"], ylim = (0,1))
    """
    dict_of_bars = {}
    for variable in variables:
        # for each variable
        variable_specific_dict_of_plots = {}
        # create plot for each df using that variable

        # Indicate if we have only self and target (without unadjusted)
        # since in this case the color of self should be red, since it's likely unadjusted.
        only_self_and_target = set(dict_of_dfs.keys()) == {"self", "target"}

        # filter dict_of_dfs
        for name in dict_of_dfs:
            df_plot_data = relative_frequency_table(
                dict_of_dfs[name],
                variable,
                dict_of_dfs[name]["weight"]
                if "weight" in dict_of_dfs[name].columns
                else None,
            )

            variable_specific_dict_of_plots[name] = go.Bar(
                x=list(df_plot_data[variable]),
                y=list(df_plot_data["prop"]),
                marker={
                    # pyre-ignore[6]: it cannot get to this point if name=="target".
                    "color": _plotly_marker_color(name, only_self_and_target, "color"),
                    "line": {
                        "color": _plotly_marker_color(
                            # pyre-ignore[6]: it cannot get to this point if name=="target".
                            name,
                            only_self_and_target,
                            "line",
                        ),
                        "width": 1.5,
                    },
                },
                opacity=0.6,
                name=naming_legend(name, list(dict_of_dfs.keys())),
                visible=True,
            )
        data = [variable_specific_dict_of_plots[name] for name in dict_of_dfs]

        layout = go.Layout(
            title=f"Sample Vs Target {variable}",
            paper_bgcolor="rgb(255, 255, 255)",
            plot_bgcolor="rgb(255, 255, 255)",
            xaxis={"title": variable},
            yaxis={
                "title": "Proportion of Total",
                "range": ylim if ylim is not None else None,
            },
        )

        fig = go.Figure(data=data, layout=layout)
        dict_of_bars[variable] = fig
        if plot_it:
            offline.iplot(fig)
    if return_dict_of_figures:
        return dict_of_bars


# TODO: add more plots other than qq for numeric (e.g.: hist, ecdf,)
# see https://plotly.com/python/distplot/
# Notice that these plots do not support the 'weight' column, so it requires a different approach.
# See the plotly_plot_density solution that uses seaborn's output
def plotly_plot_dist(
    dict_of_dfs: Dict[str, pd.DataFrame],
    variables: Optional[List[str]] = None,
    numeric_n_values_threshold: int = 15,
    weighted: bool = True,
    dist_type: Optional[Literal["kde", "qq"]] = None,
    plot_it: bool = True,
    return_dict_of_figures: bool = False,
    ylim: Optional[Tuple[float, float]] = None,
) -> Optional[Dict[str, go.Figure]]:
    """
    Plots interactive distribution plots (qq and bar plots) of the given variables.

    The plots compare the weighted distributions of an arbitrary number
    of variables from an arbitrary number of DataFrames.
    Numeric variables are plotted as either qq's using :func:`plotly_plot_qq`, or as kde desnity plots using :func:`plotly_plot_density`.
    categorical variables as barplots using :func:`plotly_plot_bar`.

    Args:
        dict_of_dfs (Dict[str, pd.DataFrame]): The key is the name of the DataFrame (E.g.: self, unadjusted, target),
            and the value is the DataFrame that contains the variables that we want to plot.
        variables (Optional[List[str]], optional): a list of variables to use for plotting. Defaults (i.e.: if None) is to use the list of all variables.
        numeric_n_values_threshold (int, optional): How many numbers should be in a column so that it is considered to be a "category"? Defaults to 15.
        weighted (bool, optional): If to use the weights with the plots. Defaults to True.
        dist_type (Optional[Literal["kde", "qq"]], optional): The type of plot to draw (relevant only for numerical variables). Defaults to None (which fallbacks to "kde").
        plot_it (bool, optional): If to plot the plots interactively instead of returning a dictionary. Defaults to True.
        return_dict_of_figures (bool, optional): If to return the dictionary containing the plots rather than just returning None. Defaults to False.
            If returned - the dictionary is of plots.
            Keys in this dictionary are the variable names for each plot.
            Values are plotly plot objects plotted like:
                offline.iplot(dict_of_all_plots['age'])
            Or simply:
                dict_of_all_plots['age']
        ylim (Optional[Tuple[float, float]], optional): A tuple with two float values representing the lower and upper limits of the y-axis.
            If not provided, the y-axis range is determined automatically. Defaults to None.
            passed to bar plots only.

    Returns:
        Optional[Dict[str, go.Figure]]: Dictionary containing plots if return_dict_of_figures is True. None otherwise.

    Examples:
        ::

            import numpy as np
            import pandas as pd
            from numpy import random
            from balance.stats_and_plots.weighted_comparisons_plots import plotly_plot_dist

            random.seed(96483)

            df = pd.DataFrame({
                'v1': random.random_integers(11111, 11114, size=100).astype(str),
                'v2': random.normal(size = 100),
                'v3': random.uniform(size = 100),
            }).sort_values(by=['v2'])

            dict_of_dfs = {
                "self": pd.concat([df, pd.Series(random.random(size = 100) + 0.5, name = "weight")], axis = 1),
                "unadjusted": pd.concat([df, pd.Series(np.ones(99).tolist() + [1000], name = "weight")], axis = 1),
                "target": pd.concat([df, pd.Series(np.ones(100), name = "weight")], axis = 1),
            }

            plotly_plot_dist(dict_of_dfs)

            # Make sure the bar plot is plotted with y in the range of 0 to 1.
            plotly_plot_dist(dict_of_dfs, ylim = (0,1))

            # See the qqplots version
            plotly_plot_dist(dict_of_dfs, dist_type="qq")
    """
    dict_of_all_plots = {}
    #  Choose set of variables to plot
    variables = choose_variables(
        *(dict_of_dfs[name] for name in dict_of_dfs), variables=variables
    )

    variables = [v for v in variables if v != "weight"]

    logger.debug(f"plotting variables {variables}")

    # TODO: patch choose_variables to return outcome_types from multiple_objects
    # find numeric values, using 'sample' df. If for some reason sample is
    # not an option, use a random df.
    if "sample" in dict_of_dfs:
        numeric_variables = (
            dict_of_dfs["sample"].select_dtypes(exclude=["object"]).columns.values
        )
    else:
        numeric_variables = (
            dict_of_dfs[random.choice(list(dict_of_dfs.keys()))]
            .select_dtypes(exclude=["object"])
            .columns.values
        )

    for _, o in enumerate(variables):
        #  Find the maximum number of non-missing values of this variable accross
        #  all the dataframes

        # Look at the first element in the dict: (name, type and values)
        logger.debug(list(dict_of_dfs.keys())[0])
        logger.debug(type(dict_of_dfs[list(dict_of_dfs.keys())[0]].loc[:, o].values))
        logger.debug(dict_of_dfs[list(dict_of_dfs.keys())[0]].loc[:, o].values)

        n_values = max(
            len(set(rm_mutual_nas(dict_of_dfs[name].loc[:, o].values)))
            for name in dict_of_dfs
        )

        if n_values == 0:
            logger.warning(f"No nonmissing values for variable '{o}', skipping")

            continue

        #  Plot categorical variables as histogram
        categorical = (o not in numeric_variables) or (
            n_values < numeric_n_values_threshold
        )

        if (dist_type is None) or dist_type == "kde":
            plotly_numeric_plot = plotly_plot_density
        elif dist_type == "qq":
            plotly_numeric_plot = plotly_plot_qq
        else:
            raise NotImplementedError(
                f"dist_type of type {dist_type} is not implemented."
            )

        # the below functions will create plotly plots
        if categorical:
            dict_of_plot = plotly_plot_bar(
                dict_of_dfs, [o], plot_it, return_dict_of_figures, ylim=ylim
            )
        else:
            # plotly_plot_density
            dict_of_plot = plotly_numeric_plot(
                dict_of_dfs, [o], plot_it, return_dict_of_figures
            )
        # the below functions will add the plotly dict outputs
        # to the dictionary 'dict_of_all_plots' (if return_dict_of_figures is True).
        if dict_of_plot is not None and return_dict_of_figures:
            dict_of_all_plots.update(dict_of_plot)

    if return_dict_of_figures:
        return dict_of_all_plots


def naming_legend(object_name: str, names_of_dfs: List[str]) -> str:
    """Returns a name for a legend of a plot given the other dfs.
    If one of the dfs we would like to plot is "unadjusted", it means
    that the Sample object contains the adjusted object as self.
    If not, then the self object is sample.

    Args:
        object_name (str): the name of the object to plot.
        names_of_dfs (List[str]): the names of the other dfs to plot.

    Returns:
        str: a string with the desired name

    Examples:
        ::

            naming_legend('self', ['self', 'target', 'unadjusted']) #'adjusted'
            naming_legend('unadjusted', ['self', 'target', 'unadjusted']) #'sample'
            naming_legend('self', ['self', 'target']) #'sample'
            naming_legend('other_name', ['self', 'target']) #'other_name'
    """
    if object_name in names_of_dfs:
        return {
            "unadjusted": "sample",
            "self": "adjusted" if "unadjusted" in names_of_dfs else "sample",
            "target": "population",
        }[object_name]
    else:
        return object_name


# TODO: set colors of the lines and dots in plots to be fixed by the object name
# (sample, target and adjusted sample).
# See the examples in Balance wiki Diagnostic_Plots page for the current plots
# (the original sample is green when we don't plot the adjusted and red when we are)


################################################################################
#  A master plotting function to navigate between seaborn and plotly plots
################################################################################


def plot_dist(
    dfs: List[Dict[str, Union[pd.DataFrame, pd.Series]]],
    names: Optional[List[str]] = None,
    variables: Optional[List[str]] = None,
    numeric_n_values_threshold: int = 15,
    weighted: bool = True,
    dist_type: Optional[Literal["kde", "hist", "qq", "ecdf"]] = None,
    library: Literal["plotly", "seaborn"] = "plotly",
    ylim: Optional[Tuple[float, float]] = None,
    **kwargs,
) -> Union[Union[List, np.ndarray], Dict[str, go.Figure], None]:
    """Plots the variables of a DataFrame by using either seaborn or plotly.

    If using plotly then using kde (or qq) plots for numeric variables and bar plots for categorical variables. Uses :func:`plotly_plot_dist`.
    If using seaborn then various types of plots are possible for the variables (see dist_type for details). Uses :func:`seaborn_plot_dist`

    Args:
        dfs (List[Dict[str, Union[pd.DataFrame, pd.Series]]]): a list (of length 1 or more) of dictionaries which describe the DataFrames and weights
            The structure is as follows:
            [
                {'df': pd.DataFrame(...), "weight": pd.Series(...)},
                ...
            ]
            The 'df' is a DataFrame which includes the column name that was supplied through 'column'.
            The "weight" is a pd.Series of weights that are used when aggregating the variable using :func:`relative_frequency_table`.
        names (List[str]): a list of the names of the DataFrames that are plotted. E.g.: ['adjusted', 'unadjusted', 'target']
            If None, then all DataFrames will be plotted, but only if library == "seaborn". (TODO: to remove this restriction)
        variables (Optional[List[str]], optional): a list of variables to use for plotting. Default (i.e.: if None) is to use the list of all variables.
        numeric_n_values_threshold (int, optional): How many numbers should be in a column so that it is considered to be a "category"? Defaults to 15.
        weighted (bool, optional): If to use the weights with the plots. Defaults to True.
        dist_type (Literal["kde", "hist", "qq", "ecdf"], optional): The type of plot to draw. The 'qq' and 'kde' options are available for library="plotly",
            While all options are available if using library="seaborn". Defaults to "kde".
        library (Literal["plotly", "seaborn"], optional): Whichever library to use for the plot. Defaults to "plotly".
        ylim (Optional[Tuple[float, float]], optional): A tuple with two float values representing the lower and upper limits of the y-axis.
            If not provided, the y-axis range is determined automatically. Defaults to None.
            passed to bar plots only.

    Raises:
        ValueError: if library is not in ("plotly", "seaborn").

    Returns:
        Union[Union[List, np.ndarray], Dict[str, go.Figure], None]:
            If library="plotly" then returns a dictionary containing plots if return_dict_of_figures is True. None otherwise.
            If library="seaborn" then returns None, unless return_axes is True. Then either a list or an np.array of matplotlib axis.

    Examples:
        ::

            import numpy as np
            import pandas as pd
            from numpy import random
            from balance.stats_and_plots.weighted_comparisons_plots import plotly_plot_bar

            random.seed(96483)

            df = pd.DataFrame({
                'v1': random.random_integers(11111, 11114, size=100).astype(str),
                'v2': random.normal(size = 100),
                'v3': random.uniform(size = 100),
            }).sort_values(by=['v2'])

            dfs1 = [
                {"df": df, "weight": pd.Series(random.random(size = 100) + 0.5)},
                {"df": df, "weight": pd.Series(np.ones(99).tolist() + [1000])},
                {"df": df, "weight": pd.Series(np.ones(100))},
            ]


            from balance.stats_and_plots.weighted_comparisons_plots import plot_dist

            # defaults to plotly with bar and qq plots. Returns None.
            plot_dist(dfs1, names=["self", "unadjusted", "target"])

            # Using seaborn, deafults to kde plots
            plot_dist(dfs1, names=["self", "unadjusted", "target"], library="seaborn") # like using dist_type = "kde"
            plot_dist(dfs1, names=["self", "unadjusted", "target"], library="seaborn", dist_type = "hist")
            plot_dist(dfs1, names=["self", "unadjusted", "target"], library="seaborn", dist_type = "qq")
            plot_dist(dfs1, names=["self", "unadjusted", "target"], library="seaborn", dist_type = "ecdf")

            plot_dist(dfs1, names=["self", "unadjusted", "target"], ylim = (0,1))
            plot_dist(dfs1, names=["self", "unadjusted", "target"], library="seaborn", dist_type = "qq", ylim = (0,1))
    """
    if library not in ("plotly", "seaborn"):
        raise ValueError(f"library must be either 'plotly' or 'seaborn', is {library}")

    #  Set default names for samples
    # TODO: this will work only with seaborn. Will need to change to something that also works for plotly.
    if names is None:
        names = [f"sample {i}" for i in range(1, len(dfs) + 1)]

    if library == "seaborn":
        return seaborn_plot_dist(
            dfs=dfs,
            names=names,
            variables=variables,
            numeric_n_values_threshold=numeric_n_values_threshold,
            weighted=weighted,
            dist_type=dist_type,
            ylim=ylim,
            **kwargs,
        )
    elif library == "plotly":
        dict_of_dfs = dict(
            zip(
                names,
                (
                    pd.concat((d["df"], pd.Series(d["weight"], name="weight")), axis=1)
                    for d in dfs
                ),
            )
        )

        if dist_type is not None:
            logger.warning("plotly plots ignore dist_type. Consider library='seaborn'")

        return plotly_plot_dist(
            dict_of_dfs,
            variables,
            numeric_n_values_threshold,
            weighted,
            # pyre-ignore[6]: plotly_plot_dist will raise a NotImplemented error if dist_type is not None, 'kde', or 'qq'
            dist_type=dist_type,
            ylim=ylim,
            **kwargs,
        )


# TODO: add plots to compare ASMD
